
#ifndef HeavyOSECPU_jitc_h
#define HeavyOSECPU_jitc_h

// Error flags
#define JITC_ERR_MASK			255
#define	JITC_ERR_PHASE0ONLY		256
#define JITC_ERR_REGNUM			(1 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_DST1			(2 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_OPECODE		(3 | JITC_ERR_PHASE0ONLY)
#define	JITC_ERR_LABELNUM		(4 | JITC_ERR_PHASE0ONLY)
#define	JITC_ERR_LABELREDEF		(5 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_PREFIX			(6 | JITC_ERR_PHASE0ONLY)
#define	JITC_ERR_LABELNODEF		7
#define	JITC_ERR_LABELTYP		8
#define	JITC_ERR_IDIOM			9
#define JITC_ERR_PREGNUM		(10 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_SRC1			(11 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_BADTYPE		(12 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_PREFIXFAR		(13 | JITC_ERR_PHASE0ONLY)
#define JITC_ERR_INTERNAL		99

// Byte operations
#define	jitCompPutByte1(p, c0)				*p++ = c0
#define	jitCompPutByte2(p, c0, c1)			*p++ = c0; *p++ = c1
#define	jitCompPutByte3(p, c0, c1, c2)		*p++ = c0; *p++ = c1; *p++ = c2
#define	jitCompPutByte4(p, c0, c1, c2, c3)	*p++ = c0; *p++ = c1; *p++ = c2; *p++ = c3

//
// functions (jitc internal)
//

// @jitc.c
void errorHandler(HOSECPU_RuntimeEnvironment *r);
int jitCompCmdLen(const unsigned char *src);

// @jitcx86.c
#if (JITC_ARCNUM == 0x0001)
//
// for x86-32bit
//

// 他のCPUへ移植する人へ:
// 以下の定数は最適化のためのものなので、すべて0として簡単に移植しても問題ありません
#define	jitCompA0001_USE_R3F_CMPJMP		1*1
#define	jitCompA0001_USE_R3F_IMM32		1*1
#define jitCompA0001_USE_R3F_IMM8		1*1
#define jitCompA0001_USE_R3F_INCDEC		1*1
#define	jitCompA0001_OPTIMIZE_JMP		1*1
#define	jitCompA0001_OPTIMIZE_MOV		1*1	/* 1にすると速度低下する？ */
#define	jitCompA0001_OPTIMIZE_CMP		1*1
#define	jitCompA0001_OPTIMIZE_ALIGN		4*1	/* 0-8を想定 */
#define	jitCompA0001_EBP128				128*1

struct JitCompWork {
	unsigned char *dst, *dst0;
	int err, maxLabels;
#if (jitCompA0001_USE_R3F_IMM32 != 0)
	int r3f;
#endif
	char prefix;	//CND命令の値を記録(初期値=0)
};

void jitCompPutImm32(struct JitCompWork *w, int i);
int jitCompGetImm32(const unsigned char *src);
int jitCompGetLabelNum(struct JitCompWork *w, const unsigned char *src);
void jitCompA0001_85DispN(struct JitCompWork *w, int disp, int n);
void jitCompA0001_movEbpDispReg32(struct JitCompWork *w, int disp, int reg32);
void jitCompA0001_movReg32EbpDisp(struct JitCompWork *w, int reg32, int disp);
void jitCompA0001_movEaxRxx(struct JitCompWork *w, int rxx);
void jitCompA0001_movRxxEax(struct JitCompWork *w, int rxx);
void jitCompA0001_fixPrefix(struct JitCompWork *w);
void jitCompA0001_checkCompPtr(struct JitCompWork *w, int p0, int p1);
void jitCompA000_loadRegCacheAll(struct JitCompWork *w);
void jitCompA000_storeRegCacheAll(struct JitCompWork *w);
void jitCompA000_loadRegCacheEcx(struct JitCompWork *w);
void jitCompA000_storeRegCacheEcx(struct JitCompWork *w);
void jitCompA000_loadRegCacheEdx(struct JitCompWork *w);
void jitCompA000_storeRegCacheEdx(struct JitCompWork *w);
int jitCompA000_selectRegCache(int rxx, int reg);
void jitCompA000_loadPRegCacheAll(struct JitCompWork *w);
void jitCompA000_storePRegCacheAll(struct JitCompWork *w);
int jitCompA000_selectPRegCache(int pxx, int reg);
int jitCompA000_convTyp(int t);
int jitCompA000_dataWidth(int t);
void jitCompA0001_checkType0(struct JitCompWork *w, int pxx, int typ, int ac);
void jitCompA0001_checkType(struct JitCompWork *w, int pxx, int typ, int ac);
void jitCompA0001_checkLimit(struct JitCompWork *w, int reg, int pxx);
int jitCompiler(unsigned char *dst, unsigned char *dst1, const unsigned char *src, const unsigned char *src1, const unsigned char *src0, HOSECPU_LabelListTag *label, int maxLabels, int level, int debugInfo1, int flags);
unsigned char *jitCompCallFunc(unsigned char *dst, void *func);
unsigned char *jitCompInit(unsigned char *dst);
void func3c(char *ebp, int opt, int r1, int p1, int lenR, int lenP, int r0, int p0);
void func3d(char *ebp, int opt, int r1, int p1, int lenR, int lenP, int r0, int p0);
void funcf4(char *ebp, int pxx, int typ, int len);
void funcf5(char *ebp, int pxx, int typ, int len);
void funcf6(char *ebp, int pxx, int typ, int len);
void funcf7(char *ebp, int pxx, int typ, int len);
void errHndl(HOSECPU_RuntimeEnvironment *r);
int jitc0(unsigned char **qq, unsigned char *q1, const unsigned char *p0, const unsigned char *p1, int level, HOSECPU_LabelListTag *label);
#if (USE_DEBUGGER != 0)
int dbgrGetRegNum(const char *p);
void dbgrMain(HOSECPU_RuntimeEnvironment *r);
#endif












#endif

#endif
